/*
Push-And-Pull Switching: Windows Switching Technique
Copyright (C) 2010 Quan Xu
E-mail:xuquan_sizuo@hotmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

using System;
using System.Collections.Generic;
using System.Text;

using System.Collections;
using System.Drawing;

namespace Algorithm
{
    //using line segment tree to calculate N rectangle union area

    // node class
    public class SegTree
    {
        public SegTree lchild;//left subtree
        public SegTree rchild;//right subtree
        public int left, right; //real section
        public int m;   // measure
        public int count;     // the number of line segment which covers this range
        public int lbd, rbd; // whether left/right end point has been covered
        public int lw, hi;            // [lw,hi]
    };

    public class LineSegmentTree
    {
        private const int INSERT = 1;
        private const int DELETE = -1;

        private SegTree IniTree(ref ArrayList Y, int lw, int hi) //create line segment tree
        {
            SegTree root = new SegTree();
            if (lw >= hi)
                return null;
            if (hi - lw == 1)
            {
                root.lchild = root.rchild = null;
            }
            else
            {
                root.lchild = IniTree(ref Y, lw, (lw + hi) / 2); //left subtree
                root.rchild = IniTree(ref Y, (lw + hi) / 2, hi); //right subtree
            }
            root.m = 0; //set m
            root.count = 0;
            root.lbd = root.rbd = 0;
            root.lw = lw;
            root.hi = hi;
            root.left = (int)Y[lw];
            root.right = (int)Y[hi];
            return root;
        }

        private void Add(ref int v, int b)
        {
               v += b;
               if(v < 0) 
                   v=0;
        }

        // we can sure [lw, hi] in the section root
        private void Update(SegTree root, int lw, int hi, int flag)
        {
            // leaf nodes and non-leaf nodes
            if (root.lchild != null && root.rchild != null)
            {     
                //non-leaf nodes
                if (hi <= root.rchild.left)
                    Update(root.lchild, lw, hi, flag); //left subtree
                else
                {
                    if (lw >= root.rchild.left)
                        Update(root.rchild, lw, hi, flag);  //right subtree
                    else
                    {
                        //between two line segments  
                        Update(root.lchild, lw, root.lchild.right, flag);
                        Update(root.rchild, root.rchild.left, hi, flag);
                    }
                }
                if (root.left == lw && root.right == hi)
                    Add(ref root.count, flag);
                if (root.left == lw)
                    Add(ref root.lbd, flag);
                if (root.right == hi)
                    Add(ref root.rbd, flag);
                if (root.count > 0) //entirely cover
                    root.m = root.right - root.left;
                else
                    root.m = root.lchild.m + root.rchild.m; //
            }
            else if (root.left == lw && root.right == hi)
            {
                // delete in the leaf nodes 
                Add(ref root.count, flag);
                Add(ref root.lbd, flag);
                Add(ref root.rbd, flag);
                if (root.count == 0)
                    root.m = 0;
                else
                    root.m = root.right - root.left;
            }
        }

        public struct Line
        {
            public int x, y1, y2;
            public int flag; // INSERT or DELETE
            public void Set(int _x,int _y1,int _y2,int _flag)
            { x = _x; y1 = _y1; y2 = _y2; flag = _flag; }
        };

        private class LineCmp:System.Collections.IComparer
        {
            public int Compare(Object a, Object b)
            {
                return ((Line)a).x - ((Line)b).x; 
            }
        }

        //O(nlogn) algorithm
        public int NRectangleUnion(ArrayList opRectangleSet)
        {

            int nCount = opRectangleSet.Count;
            int area = 0;
            if (nCount > 0)
            {
                SegTree root = new SegTree();
                ArrayList ArrY = new ArrayList();
                ArrayList ArrLine = new ArrayList();

                //initialization
                for (int i = 0; i < nCount; i++)
                {
                    Rectangle tmp = (Rectangle)opRectangleSet[i];

                    Line segline1 = new Line();
                    Line segline2 = new Line();
                    segline1.Set(tmp.Left, tmp.Top, tmp.Bottom, INSERT);
                    segline2.Set(tmp.Right, tmp.Top, tmp.Bottom, DELETE);
                    ArrLine.Add(segline1);
                    ArrLine.Add(segline2);

                    ArrY.Add(tmp.Top);
                    ArrY.Add(tmp.Bottom);
                }

                //sort
                ArrY.Sort();

                LineCmp linecmp = new LineCmp();
                ArrLine.Sort(linecmp);

                root = IniTree(ref ArrY, 0, 2 * nCount - 1);
                Line line0 = (Line)ArrLine[0];
                Update(root, line0.y1, line0.y2, line0.flag);

                for (int k = 1; k < 2 * nCount; k++)  
                {
                    Line leftline = (Line)ArrLine[k];
                    Line rightline = (Line)ArrLine[k - 1];
                    area += root.m * (leftline.x - rightline.x);
                    Update(root, leftline.y1, leftline.y2, leftline.flag);
                }
            }
            return area;
        }
    }
}
