using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics; 

using Common;

namespace Common
{

    public class CWndNode
    {
        IntPtr m_hWnd;
        CWndNode m_pPrev;
        CWndNode m_pNext;
        public CWndNode(IntPtr hWnd) { m_hWnd = hWnd; m_pPrev = m_pNext = null; }

        public IntPtr Handle
        {
            get
            {
                return m_hWnd;
            }
            set
            {
                m_hWnd = value;
            }
        }

        public CWndNode Prev
        {
            get
            {
                return m_pPrev;
            }
            set
            {
                m_pPrev = value;
            }
        }

        public CWndNode Next
        {
            get
            {
                return m_pNext;
            }
            set
            {
                m_pNext = value;
            }
        }

    };  

    public class WindowsOrder
    {

        public void DeleteWndLine(CWndNode pNode)
        {
            Debug.Assert(pNode != null);
            CWndNode pCur = pNode;
            while (pCur.Prev != null)
                pCur = pCur.Prev;
            while (pCur != null)
            {
                CWndNode pTemp = pCur;
                pCur = pCur.Next;
            }
        }

        public bool FindZOrder(IntPtr hWndA, IntPtr hWndB)
        {
            Debug.Assert(User32.IsWindow(hWndA) != 0);
            Debug.Assert(User32.IsWindow(hWndB) != 0);
            Debug.Assert(hWndA != hWndB);   

	        //
            IntPtr hTopA = hWndA;
            IntPtr hTopB = hWndB;

            IntPtr ha = User32.GetParent(hTopA);
            IntPtr hb = User32.GetParent(hTopB);
            while (User32.GetParent(hTopA) != IntPtr.Zero)
            {
                hTopA = User32.GetParent(hTopA);
            }
            while (User32.GetParent(hTopB) != IntPtr.Zero)
            {
                hTopB = User32.GetParent(hTopB);
            }

	        //case 1:these two windows A and B which belongs to different TopWindow
	        if(hTopA != hTopB)   
	       { 
                IntPtr hCur = User32.GetTopWindow(IntPtr.Zero);
                while (hCur != IntPtr.Zero)   
		        {   
			        if (hTopA == hCur)   
				        return true;   
			        else if(hTopB == hCur)   
				        return false;
                    hCur = User32.GetWindow(hCur, GW.HWNDNEXT);   
		        }   
	        }
	        //case 2:these two windows A and B which belongs to the same TopWindow
	        else   
	        {    
		        // build parent-child window 
		        CWndNode pWndLineA = new CWndNode(hWndA);   
		        CWndNode pWndLineB = new CWndNode(hWndB);

                IntPtr hTemp = hWndA;   
		        while(hTemp != hTopA)   
		        {
                    hTemp = User32.GetParent(hTemp);   
			        CWndNode pTempNode = new CWndNode(hTemp);   
			        pTempNode.Next = pWndLineA;   
			        pWndLineA.Prev = pTempNode;   
			        pWndLineA = pTempNode;   
		        }   

		        hTemp = hWndB;   
		        while(hTemp != hTopB)   
		        {
                    hTemp = User32.GetParent(hTemp);   
			        CWndNode pTempNode = new CWndNode(hTemp);   
			        pTempNode.Next = pWndLineB;   
			        pWndLineB.Prev = pTempNode;   
			        pWndLineB = pTempNode;   
		        }   

		        // query the same level windows
		        while(pWndLineA.Handle == pWndLineB.Handle)   
		        {
                    if (pWndLineA.Next == null || pWndLineB.Next == null)   
				        //one of windows is the parent of another window
			        {   
				        bool res = false;
                        if (pWndLineA.Next == null)  //A is the parent window of B,so B is front of A
					        res = false;   
				        else   
					        res = true;   
				        DeleteWndLine(pWndLineA);   
				        DeleteWndLine(pWndLineB);   
				        return res;   
			        }   
			        pWndLineA = pWndLineA.Next;   
			        pWndLineB = pWndLineB.Next;   
		        }

                //pWndLineA.Handle and pWndLineB.Handle is the same level   
                //pWndLineA.Handle does not always equal hWndA
                //pWndLineB.Handle does not always equal hWndB
                IntPtr hA = pWndLineA.Handle;
                IntPtr hB = pWndLineB.Handle;   
		        DeleteWndLine(pWndLineA);   
		        DeleteWndLine(pWndLineB);

                IntPtr hCur = User32.GetTopWindow(User32.GetParent(hA));
                while (hCur != IntPtr.Zero)   
		        {   
			        if(hA == hCur)   
				        return true;   
			        else if(hB == hCur)   
				        return false;
                    hCur = User32.GetWindow(hCur, GW.HWNDNEXT); 
		        }   
	        }

           Trace.WriteLine("Function error! FindZOrder appears unexpected result!");
           Debug.Assert(false);   

	        return   false; 
        }
    }
}
