/*
Push-And-Pull Switching: Windows Switching Technique
Copyright (C) 2010 Quan Xu
E-mail:xuquan_sizuo@hotmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Collections;
using System.Threading;

using Common;
using Algorithm;
using WindowsGroups;
using DesktopWindows;
using PushAndPullPGlobalHooks;

namespace PushAndPullSwitching
{
    public partial class PushAndPullSwitching : Form
    {
        //Global hook
        private GlobalHooks m_GlobalHooks;

        //Push-And-Pull--Group Switching
        private int restackingtime = 0;
        private int pushandpulltime = 0;

        //mouse wheel navigation
        private bool isaltshiftpress = false;
        private bool isaltpress = false;
        private int wheelnumber = 0;
        private int rotateddnumber = 0;
        private int lastwheelstate = 0;
        private ArrayList PPMWRestacking = new ArrayList();
        private ArrayList PPMWPushAndPull = new ArrayList();
        private int altpushandpulltime = 0;
        private int altrestackingtime = 0;

        //windows group
        private ArrayList PushAndPullGroup = new ArrayList();
        private ArrayList Restacking = new ArrayList();

        //overlapping
        //private const double overlappingratio = 0.15; 
        private  double overlappingratio; 

        public PushAndPullSwitching()
        {
            InitializeComponent();

            // Instantiate our GlobalHooks object
            m_GlobalHooks = new GlobalHooks(this.Handle);

            //WH_SHELL Hook
            m_GlobalHooks.Shell.WindowActivated += new PushAndPullPGlobalHooks.GlobalHooks.WindowEventHandler(_GlobalHooks_ShellWindowActivated);

            //WH_KEYBOARD_LL Hook
            m_GlobalHooks.KeyboardLL.KeyDownEvent += new KeyEventHandler(KeyBoardLL_KeyDown);
            m_GlobalHooks.KeyboardLL.KeyUpEvent += new KeyEventHandler(KeyBoardLL_KeyUp);

            //WH_MOUSE_LL Hook
            m_GlobalHooks.MouseLL.MouseWheel += new MouseEventHandler(MouseLL_MouseWheel);

            //Start
            m_GlobalHooks.Shell.Start();
            m_GlobalHooks.KeyboardLL.Start();
            m_GlobalHooks.MouseLL.Start();

            //
            overlappingratio = Properties.Settings.Default.overlappingratio; //read from config file
        }

        protected override void OnHandleCreated(EventArgs e)
        {
            base.OnHandleCreated(e);
        }

        protected override void WndProc(ref Message m)
        {
            // Check to see if we have received any Windows messages telling us about our hooks
            if (m_GlobalHooks != null)
            {
                m_GlobalHooks.ProcessWindowMessage(ref m);
            }
            base.WndProc(ref m);
        }

        //Hook WH_SHELL
        private void _GlobalHooks_ShellWindowActivated(IntPtr Handle)
        {
            if (Handle != IntPtr.Zero)
            {
                Desktop m_windows = Desktop.Empty;
                m_windows.SetActiveWindows(Handle);
            }
        }

        //HOOK WH_KEYBOARD_LL
        private void KeyBoardLL_KeyDown(object sender, KeyEventArgs e)
        {
            isaltshiftpress = false;
            isaltpress = false;

            #region Push-And-Pull Switching
            //Keyboard Navigation
            if (e.Control)
            {
                if (e.Shift) //Push-And-Pull -- Restacking
                {
                    if (e.KeyCode == Keys.Up)
                    {
                        restackingtime++;

                        if (restackingtime == 1) //it only calculates one time in one process
                        {
                            DivideClusterGroups(ref Restacking, overlappingratio);
                        }
                        ReStackingKeyUp(ref Restacking);
                    }
                    else if (e.KeyCode == Keys.Down)
                    {
                        ReStackingKeyDown(ref Restacking);
                    }
                }
                else //Push-And-Pull -- Group Switching -- Keyboard Navigation
                {
                    if (e.KeyCode == Keys.Up)
                    {
                        pushandpulltime++;
                        if (pushandpulltime == 1)
                        {
                            //create windows group
                            DivideWindowsGroups(ref PushAndPullGroup, overlappingratio);
                        }

                        //Push Action
                        PushGroup(ref PushAndPullGroup);
                        //FixPushGroup(ref PushAndPullGroup);
                    }
                    else if (e.KeyCode == Keys.Down)
                    {
                        //Pull Action
                        PullGroup(ref PushAndPullGroup);
                    }
                }
            }
            #endregion
            //////////////////////////////////////////////////////////////////////////

            //////////////////////////////////////////////////////////////////////////
            #region mouse wheel navigation for Push-And-Pull/Restacking
            //mouse wheel navigation
            //using Alt+mouse wheel, because Ctrl + mouse wheel was used by OS 
            //Push-And-Pull Switching--Group Switching
            if (e.Alt)
            {
                isaltpress = true;
                if (e.Shift)
                {
                    altrestackingtime++;
                    if (altrestackingtime == 1) //it only calculates one time in one process
                    {
                        DivideClusterGroups(ref PPMWRestacking, overlappingratio);
                    }
                }
                else
                {
                    altpushandpulltime++;
                    isaltshiftpress = true;
                    if (altpushandpulltime == 1)
                    {
                        DivideWindowsGroups(ref PPMWPushAndPull, overlappingratio);
                    }
                }
            }
            #endregion
            //////////////////////////////////////////////////////////////////////////
        }

        private void KeyBoardLL_KeyUp(object sender, KeyEventArgs e)
        {
            /////////////////////////////////////////////////////////////////////////
            //Push-And-Pull Switching -- Keyboard Navigation 
            //Release Ctrl
            if (e.KeyCode == Keys.RControlKey || e.KeyCode == Keys.LControlKey)
            {
                restackingtime = 0;
                if(Restacking.Count > 0)
                    ReleasePushAndPull(ref Restacking);
            
                pushandpulltime = 0;
                if (PushAndPullGroup.Count > 0)
                    ReleasePushAndPull(ref PushAndPullGroup);
                
            }
            //////////////////////////////////////////////////////////////////////////


            //////////////////////////////////////////////////////////////////////////
            //Push-And-Pull Switching -- Mouse Wheel Navigation
            //Release Alt
            if (e.KeyCode == Keys.LMenu || e.KeyCode == Keys.RMenu)
            {
                rotateddnumber = 0;
                lastwheelstate = 0;

                altrestackingtime = 0;
                if (PPMWRestacking.Count > 0)
                    ReleasePushAndPull(ref PPMWRestacking);

                altpushandpulltime = 0;
                if (PPMWPushAndPull.Count > 0)
                    ReleasePushAndPull(ref PPMWPushAndPull);
                
            }
            //////////////////////////////////////////////////////////////////////////
        }

        //Push-And-Pull Switching -- Mouse Wheel Navigation
        private void MouseLL_MouseWheel(object sender, MouseEventArgs e)
        {
            wheelnumber = e.Delta / 120;
            rotateddnumber += wheelnumber;

            if (isaltpress)
            {
                if (isaltshiftpress)
                {
                    MouseWheelPushAndPull(ref PPMWRestacking, rotateddnumber);
                }
                else
                {
                    MouseWheelPushAndPull(ref PPMWPushAndPull, rotateddnumber);
                }
            }
        }

        //create windows group (support multi-screen)
        private void DivideWindowsGroups(ref ArrayList resGroups, double overlappingratio)
        {
            WindowsGroup m_WindowGroup = new WindowsGroup();
            Desktop m_windows = Desktop.DesktopWindows;
            if (m_windows != null)
            {
                if (m_windows.ForegroundWindow != null)
                {
                    resGroups.Clear();

                    Cluster cluster = new Cluster();
                    ArrayList ScreenWindows = new ArrayList();
                    cluster.DivideWindowsByScreen(Screen.AllScreens, ref ScreenWindows);

                    int curScreenIndex = m_WindowGroup.GetDefinitionActiveWindowGroupIndex(m_windows.ForegroundWindow.Handle, ScreenWindows);
                    if (curScreenIndex >= 0 && curScreenIndex < ScreenWindows.Count)
                    {
                        ArrayList oriWindows = (ArrayList)ScreenWindows[curScreenIndex];
                        m_WindowGroup.DivideGroups(m_windows.ForegroundWindow, oriWindows, resGroups, overlappingratio);
                    }
                }
            }
        }

        //create cluster
        private void DivideClusterGroups(ref ArrayList resGroups, double overlappingratio)
        {
            Cluster cluster = new Cluster();
            ArrayList windowsclusters = new ArrayList();
            cluster.DivideCluster(ref windowsclusters);

            int nCount = windowsclusters.Count;
            if (nCount > 0)
            {
                ArrayList currentcluster = new ArrayList();
                currentcluster = (ArrayList)windowsclusters[0]; // the current cluster

                if (currentcluster.Count > 1)
                {
                    Window forewindow = (Window)currentcluster[0];
                    resGroups.Clear();
                    WindowsGroup m_WindowsLayer = new WindowsGroup();
                    m_WindowsLayer.DivideGroups(forewindow, currentcluster, resGroups, overlappingratio);

                    //only keep active window
                    ArrayList activeWindowsList = ((ArrayList)resGroups[0]);
                    int activeGroupNumber = activeWindowsList.Count;
                    if (activeGroupNumber > 1) 
                    {
                        activeWindowsList.RemoveRange(1, activeGroupNumber - 1);
                    }
                }
            }
        }

        #region Push-And-Pull -- Restacking
        /// <summary>
        /// Restacking Key Up
        /// </summary>
        /// <param name="oplayers"></param>
        private void ReStackingKeyUp(ref ArrayList oplayers)
        {
            Desktop m_windows = Desktop.DesktopWindows;
            Window m_window = m_windows.GetDefinitionActiveWindow();

            WindowsGroup m_WindowsLayer = new WindowsGroup();
            int DefinitionActiveLayerIndex = -1;
            if (m_window != null)
            {
                DefinitionActiveLayerIndex = m_WindowsLayer.GetDefinitionActiveWindowGroupIndex(m_window.Handle, oplayers);
            }
            else
            {
                if (oplayers.Count > 0)
                {
                    ArrayList firstGruop = new ArrayList();
                    firstGruop = (ArrayList)oplayers[0];
                    m_windows.SetActiveWindows(((Window)firstGruop[0]).Handle);
                }
                DefinitionActiveLayerIndex = 0;
            }

            int groupNumber = oplayers.Count;
            if (groupNumber > 0)
            {
                if (DefinitionActiveLayerIndex > -1 && DefinitionActiveLayerIndex < oplayers.Count - 1)
                {
                    ArrayList activeWindowsList = ((ArrayList)oplayers[DefinitionActiveLayerIndex]);

                    //int activeGroupNumber = activeWindowsList.Count;
                    //if (activeGroupNumber > 1) //only keep active window
                    //{
                    //    activeWindowsList.RemoveRange(1, activeGroupNumber - 1);
                    //}

                    ArrayList beforeWindowsList = ((ArrayList)oplayers[DefinitionActiveLayerIndex + 1]);
                    int beforeNumber = beforeWindowsList.Count;

                    Window before = (Window)beforeWindowsList[beforeNumber - 1];

                    Window temp = (Window)activeWindowsList[0];
                    User32.SetWindowPos(temp.Handle, before.Handle, temp.Left, temp.Top, temp.Width, temp.Height, Constants.SWP_NOACTIVATE); //& (Constants.SWP_NOMOVE & Constants.SWP_NOSIZE)

                    m_WindowsLayer.SwitchGroup(DefinitionActiveLayerIndex, DefinitionActiveLayerIndex + 1, ref oplayers);
                }
            }
        }

        /// <summary>
        /// Restacking Key Down
        /// </summary>
        /// <param name="oplayers"></param>
        private void ReStackingKeyDown(ref ArrayList oplayers)
        {
            Desktop m_windows = Desktop.DesktopWindows;
            Window m_window = m_windows.GetDefinitionActiveWindow();

            WindowsGroup m_WindowsLayer = new WindowsGroup();
            int DefinitionActiveLayerIndex = -1;
            if (m_window != null)
            {
                DefinitionActiveLayerIndex = m_WindowsLayer.GetDefinitionActiveWindowGroupIndex(m_window.Handle, oplayers);
            }
            else
            {
                if (oplayers.Count > 0)
                {
                    ArrayList firstGruop = new ArrayList();
                    firstGruop = (ArrayList)oplayers[0];
                    m_windows.SetActiveWindows(((Window)firstGruop[0]).Handle);
                }
                DefinitionActiveLayerIndex = 0;
            }

            int groupNumber = oplayers.Count;
            if (groupNumber > 0)
            {
                if (DefinitionActiveLayerIndex > 0 && DefinitionActiveLayerIndex <= oplayers.Count - 1)
                {
                    ArrayList activeWindowsList = ((ArrayList)oplayers[DefinitionActiveLayerIndex]);
                    int activeGroupNumber = activeWindowsList.Count;

                    if (DefinitionActiveLayerIndex == 1)
                    {
                        Window before = (Window)activeWindowsList[0];
                        User32.SetWindowPos(before.Handle, (IntPtr)Constants.HWND_TOP, before.Left, before.Top, before.Width, before.Height, Constants.SWP_NOACTIVATE & (Constants.SWP_NOMOVE & Constants.SWP_NOSIZE));
                    }
                    else
                    {
                        ArrayList beforeWindowsList = ((ArrayList)oplayers[DefinitionActiveLayerIndex - 2]);
                        int beforeNumber = beforeWindowsList.Count;

                        Window before = (Window)beforeWindowsList[beforeNumber - 1];
                        Window temp = (Window)activeWindowsList[0];
                        User32.SetWindowPos(temp.Handle, before.Handle, temp.Left, temp.Top, temp.Width, temp.Height, Constants.SWP_NOACTIVATE);
                    }

                    m_WindowsLayer.SwitchGroup(DefinitionActiveLayerIndex, DefinitionActiveLayerIndex - 1, ref oplayers);
                }
            }
        }

        #endregion

        #region Push-And-Pull -- Group Switching -- Keyboard Navigation
        //Push Group Action
        private void PushGroup(ref ArrayList opGroup)
        {
            Desktop m_windows = Desktop.DesktopWindows;
            if (m_windows != null)
            {
                WindowsGroup m_WindowGroup = new WindowsGroup();

                Window acwindow = m_windows.GetDefinitionActiveWindow();
                int DefinitionActiveLayerIndex = -1;
                if (acwindow != null)
                {
                    DefinitionActiveLayerIndex = m_WindowGroup.GetDefinitionActiveWindowGroupIndex(acwindow.Handle, opGroup);
                }
                else
                {
                    if (opGroup.Count > 0)
                    {
                        ArrayList firstGruop = new ArrayList();
                        firstGruop = (ArrayList)opGroup[0];
                        m_windows.SetActiveWindows(((Window)firstGruop[0]).Handle);
                    }
                    DefinitionActiveLayerIndex = 0;
                }

                if (opGroup.Count > 1)
                {
                    if (DefinitionActiveLayerIndex > -1 && DefinitionActiveLayerIndex < opGroup.Count - 1)
                    {
                        ArrayList activeWindowsList = ((ArrayList)opGroup[DefinitionActiveLayerIndex]);
                        int activeGroupNumber = activeWindowsList.Count;

                        ArrayList beforeWindowsList = ((ArrayList)opGroup[DefinitionActiveLayerIndex + 1]);
                        int beforeNumber = beforeWindowsList.Count;

                        Window before = (Window)beforeWindowsList[beforeNumber - 1];

                        Window temp = (Window)activeWindowsList[0];
                        User32.SetWindowPos(temp.Handle, before.Handle, temp.Left, temp.Top, temp.Width, temp.Height, Constants.SWP_NOACTIVATE); //& (Constants.SWP_NOMOVE & Constants.SWP_NOSIZE)
                        if (activeGroupNumber > 1)
                        {
                            for (int i = 1; i < activeGroupNumber; i++)
                            {
                                temp = (Window)activeWindowsList[i];
                                Window tempBefore = (Window)activeWindowsList[i - 1];
                                User32.SetWindowPos(temp.Handle, tempBefore.Handle, temp.Left, temp.Top, temp.Width, temp.Height, Constants.SWP_NOACTIVATE); //& (Constants.SWP_NOMOVE & Constants.SWP_NOSIZE)
                            }
                        }
                        m_WindowGroup.SwitchGroup(DefinitionActiveLayerIndex, DefinitionActiveLayerIndex + 1, ref opGroup);
                    }
                }
            }
        }

        //Pull Group Action
        private void PullGroup(ref ArrayList opGroup)
        {
            Desktop m_windows = Desktop.DesktopWindows;
            if (m_windows != null)
            {
                WindowsGroup m_WindowGroup = new WindowsGroup();

                Window acwindow = m_windows.GetDefinitionActiveWindow();
                int DefinitionActiveLayerIndex = -1;
                if (acwindow != null)
                {
                    DefinitionActiveLayerIndex = m_WindowGroup.GetDefinitionActiveWindowGroupIndex(acwindow.Handle, opGroup);
                }
                else
                {
                    if (opGroup.Count > 0)
                    {
                        ArrayList firstGruop = new ArrayList();
                        firstGruop = (ArrayList)opGroup[0];
                        m_windows.SetActiveWindows(((Window)firstGruop[0]).Handle);
                    }
                    DefinitionActiveLayerIndex = 0;
                }

                if (opGroup.Count > 1)
                {
                    if (DefinitionActiveLayerIndex > 0 && DefinitionActiveLayerIndex <= opGroup.Count - 1)
                    {
                        ArrayList activeWindowsList = ((ArrayList)opGroup[DefinitionActiveLayerIndex]);
                        int activeGroupNumber = activeWindowsList.Count;

                        if (DefinitionActiveLayerIndex == 1)
                        {
                            Window before = (Window)activeWindowsList[0];

                            User32.SetWindowPos(before.Handle, (IntPtr)Constants.HWND_TOP, before.Left, before.Top, before.Width, before.Height, Constants.SWP_NOACTIVATE & (Constants.SWP_NOMOVE & Constants.SWP_NOSIZE));
                            if (activeGroupNumber > 1)
                            {
                                for (int i = 1; i < activeGroupNumber; i++)
                                {
                                    Window temp = (Window)activeWindowsList[i];
                                    Window tempBefore = (Window)activeWindowsList[i - 1];
                                    User32.SetWindowPos(temp.Handle, tempBefore.Handle, temp.Left, temp.Top, temp.Width, temp.Height, Constants.SWP_NOACTIVATE);
                                }
                            }
                        }
                        else
                        {
                            ArrayList beforeWindowsList = ((ArrayList)opGroup[DefinitionActiveLayerIndex - 2]);
                            int beforeNumber = beforeWindowsList.Count;

                            Window before = (Window)beforeWindowsList[beforeNumber - 1];
                            Window temp = (Window)activeWindowsList[0];
                            User32.SetWindowPos(temp.Handle, before.Handle, temp.Left, temp.Top, temp.Width, temp.Height, Constants.SWP_NOACTIVATE); //& (Constants.SWP_NOMOVE & Constants.SWP_NOSIZE)

                            for (int i = 1; i < activeGroupNumber; i++)
                            {
                                temp = (Window)activeWindowsList[i];
                                before = (Window)activeWindowsList[i - 1];
                                User32.SetWindowPos(temp.Handle, before.Handle, temp.Left, temp.Top, temp.Width, temp.Height, Constants.SWP_NOACTIVATE);
                            }
                        }
                        m_WindowGroup.SwitchGroup(DefinitionActiveLayerIndex, DefinitionActiveLayerIndex - 1, ref opGroup);
                    }
                }
            }
        }

        #endregion

        #region Push-And-Pull -- Group Switching -- Keyboard Navigation (Another way)
        //some windows. such as excel, vs2005, they do not work very well, it may be produced by SetWindowPos function (parameter),so I try to use the parameter(HWND_NOTOPMOST,HWND_TOPMOST,HWND_TOP,HWND_BOTTOM) to do. 
        private void FixPushGroup(ref ArrayList opGroup)                                                                                                             
        {                                                                                                                                                        
            Desktop m_windows = Desktop.DesktopWindows;                                                                                                         
            if (m_windows != null)
            {
                WindowsGroup m_WindowGroup = new WindowsGroup();

                Window acwindow = m_windows.GetDefinitionActiveWindow();
                int DefinitionActiveLayerIndex = -1;
                if (acwindow != null)
                {
                    DefinitionActiveLayerIndex = m_WindowGroup.GetDefinitionActiveWindowGroupIndex(acwindow.Handle, opGroup);
                }
                else
                {
                    if (opGroup.Count > 0)
                    {
                        ArrayList firstGruop = new ArrayList();
                        firstGruop = (ArrayList)opGroup[0];
                        m_windows.SetActiveWindows(((Window)firstGruop[0]).Handle);
                    }
                    DefinitionActiveLayerIndex = 0;
                }

                if (opGroup.Count > 1)
                {
                    if (DefinitionActiveLayerIndex > -1 && DefinitionActiveLayerIndex < opGroup.Count - 1)
                    {
                        ArrayList beforeWindowsList = ((ArrayList)opGroup[DefinitionActiveLayerIndex + 1]);
                        int beforeNumber = beforeWindowsList.Count;
                        for (int i = beforeNumber - 1; i >= 0; i--)
                        {
                            Window temp = (Window)beforeWindowsList[i];
                            User32.SetWindowPos(temp.Handle, (IntPtr)Constants.HWND_TOP, temp.Left, temp.Top, temp.Width, temp.Height, Constants.SWP_NOACTIVATE);
                        }
                        
                        if(DefinitionActiveLayerIndex > 0)
                        {
                            for (int j = DefinitionActiveLayerIndex - 1; j >= 0; j--) 
                            {
                                ArrayList tmpList = ((ArrayList)opGroup[j]);
                                for (int k = tmpList.Count - 1; k >= 0; k--) 
                                {
                                    Window temp = (Window)tmpList[k];
                                    User32.SetWindowPos(temp.Handle, (IntPtr)Constants.HWND_TOP, temp.Left, temp.Top, temp.Width, temp.Height, Constants.SWP_NOACTIVATE);
                                }
                            }
                        }

                        m_WindowGroup.SwitchGroup(DefinitionActiveLayerIndex, DefinitionActiveLayerIndex + 1, ref opGroup);
                    }
                }
            }
        }
        #endregion

        #region Push-And-Pull -- Mouse Wheel Navigation
        private void MouseWheelPushAndPull(ref ArrayList oplayers, int numberrotatedwheel)
        {
            Desktop m_windows = Desktop.DesktopWindows;
            Window m_window = m_windows.GetDefinitionActiveWindow();

            WindowsGroup m_WindowsLayer = new WindowsGroup();
            int DefinitionActiveLayerIndex = -1;
            if (m_window != null)
            {
                DefinitionActiveLayerIndex = m_WindowsLayer.GetDefinitionActiveWindowGroupIndex(m_window.Handle, oplayers);
            }
            else
            {
                if (oplayers.Count > 0)
                {
                    ArrayList firstGruop = new ArrayList();
                    firstGruop = (ArrayList)oplayers[0];
                    if(firstGruop.Count > 0)
                        m_windows.SetActiveWindows(((Window)firstGruop[0]).Handle);
                }
                DefinitionActiveLayerIndex = 0;
            }

            int groupNumber = oplayers.Count;

            if (groupNumber > 0)
            {
                //using cycle pattern
                int objindex = numberrotatedwheel % groupNumber;
                if((DefinitionActiveLayerIndex >= 0) && (DefinitionActiveLayerIndex < oplayers.Count))
                {
                    ArrayList activeWindowsList = ((ArrayList)oplayers[DefinitionActiveLayerIndex]);
                    int activeGroupNumber = activeWindowsList.Count;

                    if (numberrotatedwheel > lastwheelstate)
                    {
                        if (DefinitionActiveLayerIndex == oplayers.Count - 1)
                        {
                            Window temp = (Window)activeWindowsList[0];
                            User32.SetWindowPos(temp.Handle, (IntPtr)Constants.HWND_TOP, temp.Left, temp.Top, temp.Width, temp.Height, Constants.SWP_NOACTIVATE); //& (Constants.SWP_NOMOVE & Constants.SWP_NOSIZE)
                            if (activeGroupNumber > 1)
                            {
                                for (int i = 1; i < activeGroupNumber; i++)
                                {
                                    temp = (Window)activeWindowsList[i];
                                    Window tempBefore = (Window)activeWindowsList[i - 1];
                                    User32.SetWindowPos(temp.Handle, tempBefore.Handle, temp.Left, temp.Top, temp.Width, temp.Height, Constants.SWP_NOACTIVATE); //& (Constants.SWP_NOMOVE & Constants.SWP_NOSIZE)
                                }
                            }
                        }
                        else if (DefinitionActiveLayerIndex > -1 && DefinitionActiveLayerIndex < oplayers.Count - 1)
                        {
                            ArrayList beforeWindowsList = ((ArrayList)oplayers[DefinitionActiveLayerIndex + 1]);
                            int beforeNumber = beforeWindowsList.Count;

                            Window before = (Window)beforeWindowsList[beforeNumber - 1];

                            Window temp = (Window)activeWindowsList[0];
                            User32.SetWindowPos(temp.Handle, before.Handle, temp.Left, temp.Top, temp.Width, temp.Height, Constants.SWP_NOACTIVATE); //& (Constants.SWP_NOMOVE & Constants.SWP_NOSIZE)
                            if (activeGroupNumber > 1)
                            {
                                for (int i = 1; i < activeGroupNumber; i++)
                                {
                                    temp = (Window)activeWindowsList[i];
                                    Window tempBefore = (Window)activeWindowsList[i - 1];
                                    User32.SetWindowPos(temp.Handle, tempBefore.Handle, temp.Left, temp.Top, temp.Width, temp.Height, Constants.SWP_NOACTIVATE); //& (Constants.SWP_NOMOVE & Constants.SWP_NOSIZE)
                                }
                            }
                        }
                        m_WindowsLayer.SwitchGroup(DefinitionActiveLayerIndex, (DefinitionActiveLayerIndex + 1) % groupNumber, ref oplayers);
                    }
                    else if (numberrotatedwheel < lastwheelstate)
                    {
                        if (DefinitionActiveLayerIndex == 0)
                        {
                            ArrayList beforeWindowsList = ((ArrayList)oplayers[groupNumber - 1]);
                            int beforeNumber = beforeWindowsList.Count;

                            Window before = (Window)beforeWindowsList[beforeNumber - 1];
                            Window temp = (Window)activeWindowsList[0];
                            User32.SetWindowPos(temp.Handle, before.Handle, temp.Left, temp.Top, temp.Width, temp.Height, Constants.SWP_NOACTIVATE); //& (Constants.SWP_NOMOVE & Constants.SWP_NOSIZE)

                            for (int i = 1; i < activeGroupNumber; i++)
                            {
                                temp = (Window)activeWindowsList[i];
                                before = (Window)activeWindowsList[i - 1];
                                User32.SetWindowPos(temp.Handle, before.Handle, temp.Left, temp.Top, temp.Width, temp.Height, Constants.SWP_NOACTIVATE);
                            }
                        }
                        else if (DefinitionActiveLayerIndex == 1)
                        {
                            Window before = (Window)activeWindowsList[0];

                            User32.SetWindowPos(before.Handle, (IntPtr)Constants.HWND_TOP, before.Left, before.Top, before.Width, before.Height, Constants.SWP_NOACTIVATE & (Constants.SWP_NOMOVE & Constants.SWP_NOSIZE));
                            if (activeGroupNumber > 1)
                            {
                                for (int i = 1; i < activeGroupNumber; i++)
                                {
                                    Window temp = (Window)activeWindowsList[i];
                                    Window tempBefore = (Window)activeWindowsList[i - 1];
                                    User32.SetWindowPos(temp.Handle, tempBefore.Handle, temp.Left, temp.Top, temp.Width, temp.Height, Constants.SWP_NOACTIVATE);
                                }
                            }
                        }
                        else if (DefinitionActiveLayerIndex > 1 && DefinitionActiveLayerIndex <= oplayers.Count - 1)
                        {
                            ArrayList beforeWindowsList = ((ArrayList)oplayers[DefinitionActiveLayerIndex - 2]);
                            int beforeNumber = beforeWindowsList.Count;

                            Window before = (Window)beforeWindowsList[beforeNumber - 1];
                            Window temp = (Window)activeWindowsList[0];
                            User32.SetWindowPos(temp.Handle, before.Handle, temp.Left, temp.Top, temp.Width, temp.Height, Constants.SWP_NOACTIVATE); //& (Constants.SWP_NOMOVE & Constants.SWP_NOSIZE)

                            for (int i = 1; i < activeGroupNumber; i++)
                            {
                                temp = (Window)activeWindowsList[i];
                                before = (Window)activeWindowsList[i - 1];
                                User32.SetWindowPos(temp.Handle, before.Handle, temp.Left, temp.Top, temp.Width, temp.Height, Constants.SWP_NOACTIVATE);
                            }
                        }

                        m_WindowsLayer.SwitchGroup(DefinitionActiveLayerIndex, (DefinitionActiveLayerIndex - 1 + groupNumber) % groupNumber, ref oplayers);
                    }
                    lastwheelstate = numberrotatedwheel;
                }
            }
        }

        #endregion

        //release keyboard/mouse wheel
        private void ReleasePushAndPull(ref ArrayList oplayers)
        {
            if (oplayers.Count > 0)
            {
                ArrayList firstGruop = new ArrayList();
                firstGruop = (ArrayList)oplayers[0];

                Window forewindow = (Window)firstGruop[0];
        
                Desktop m_windows = Desktop.DesktopWindows;
                if (m_windows != null && forewindow != null)
                {
                    if(m_windows.Count > 0)
                    {
                        if (m_windows.ForegroundWindow != null)
                        {
                            if (m_windows.ForegroundWindow.Handle != forewindow.Handle)
                            {
                                IntPtr hForeWnd = m_windows.ForegroundWindow.Handle;
                                User32.AllowSetForegroundWindow(User32.ASFW_ANY);
                                Thread.Sleep(100);
                                SendKeys.Send("^"); //ctrl
                                User32.SetForegroundWindow(forewindow.Handle);
                                User32.SwitchToThisWindow(forewindow.Handle, true);
                                m_windows.SetActiveWindows(forewindow.Handle);
                            }
                        }
                    }
                }
            }
        }

        //exit application
        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (MessageBox.Show("Are you sure you want to exit Push-And-Pull Switching?", "Push-And-Pull Switching", MessageBoxButtons.OKCancel, MessageBoxIcon.None) == DialogResult.OK)
            {
                this.PushAndPullnotifyIcon.Visible = false;

                m_GlobalHooks.Shell.Stop();
                m_GlobalHooks.KeyboardLL.Stop();
                m_GlobalHooks.MouseLL.Stop();

                Application.Exit();
            }
        }

        private void PushAndPullSwitching_Load(object sender, EventArgs e)
        {
            this.Hide();
        }
    }
}