using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using System.Diagnostics;

using Common;

namespace DesktopWindows
{
    public class WindowsApp:Dictionary<int,Windows>
    {
         private bool EnumWindowCallBack(IntPtr hWnd, int lParam)
        {
            uint z_order = (uint)this.Count;
            Windows window = new Windows(hWnd, z_order, false);
            bool skip = false;

            if (window.WindowName.Length == 0)
            {
                skip = true;
            }
            else if (lParam == 1 && window.Visible == false) {
                skip = true;
            }

            if (!skip)
            {
                this.Add(hWnd.ToInt32(), new Windows(hWnd, z_order, false));
            }
            return true;
        }

        public static WindowsApp Empty
        {
            get { return new WindowsApp(); }
        }

        public static WindowsApp All
        {
            get
            {
                WindowsApp windows = new WindowsApp();
                windows.Refresh(false);
                return windows;
            }
        }

        public static WindowsApp Visible
        {
            get
            {
                WindowsApp windows = new WindowsApp();
                windows.Refresh(true);
                return windows;
            }
        }

        public WindowsApp()
            : base()
        {
        }

        public void Refresh(bool visibleOnly)
        {
            Clear();
            User32.EnumWindows(this.EnumWindowCallBack, visibleOnly ? 1 : 0);
        }

        public Windows ForegroundWindow
        {
            get
            {
                IntPtr h = User32.GetForegroundWindow();
                if (this.ContainsKey(h.ToInt32()))
                {
                    return this[h.ToInt32()];
                }
                else
                {
                    return null;
                }
            }

            set
            {
                User32.SwitchToThisWindow(value.Handle, false);
            }
        }

        public void DisplayAllTopWindows(IntPtr Hwnd, Int32 level, ArrayList m_AllTopWindowsHandle)
        {
            //display all top windows

            Hwnd = User32.GetTopWindow(Hwnd);
            if (Hwnd == IntPtr.Zero)
            {
                Debug.Assert(false);
                return;
            }
            while (Hwnd != IntPtr.Zero)
            {
                if (User32.IsWindowVisible(Hwnd))
                {
                    Int64 l = User32.GetWindowLong(Hwnd, WindowStyles.GWL_STYLE);
                    if (((l & WindowStyles.WS_VISIBLE) != 0) && ((l & WindowStyles.WS_OVERLAPPEDWINDOW) != 0))
                    {
                        m_AllTopWindowsHandle.Add(Hwnd);
                    }
                }

                Hwnd = User32.GetWindow(Hwnd, GW.HWNDNEXT);
            }
        }

        public void GetWindowOrder(ArrayList WindowsHandle)
        {
            WindowsOrder m_WindowOrder = new WindowsOrder();
            int nSize = WindowsHandle.Count;
            if (nSize > 0)
            {
                IntPtr hWnd;
                for (int i = 0; i < nSize - 1; i++)
                {
                    for (int j = 0; j < nSize - 1 - i; j++)
                    {
                        if (!m_WindowOrder.FindZOrder((IntPtr)WindowsHandle[j], (IntPtr)WindowsHandle[j + 1]))
                        {
                            hWnd = (IntPtr)WindowsHandle[j];
                            WindowsHandle[j] = WindowsHandle[j + 1];
                            WindowsHandle[j + 1] = hWnd;
                        }
                    }
                }
            }
        }

        //bring hWnd to the top
        public void ChangeWindowOrder(IntPtr hWnd)
        {
            bool isTop = User32.BringWindowToTop(hWnd); //SetForegroundWindow,SetWindowPos
            if (!isTop)
            {
                Console.WriteLine("Bring the window to top is failed.");
            }
        }
    }
}
