using System;
using System.Collections.Generic;
using System.Text;

using Common;

namespace NovelTechnique
{
    public class CDeferPos
    {
        private IntPtr m_hdwp;

        // Constructor
        // This sets up the RAIA idiom by calling BeginDeferWindowPos. The number of windows
        // can be passed as an argument to optimize memory management, although the API will
        // grow the memory if needed at run time.
        CDeferPos(int nWindows)
        {
            m_hdwp = User32.BeginDeferWindowPos(nWindows);
        }

        public bool Release()
        {
            return User32.EndDeferWindowPos(m_hdwp);
        }

        // MoveWindow
        // Emulates a call to ::MoveWindow but the actual call is delayed until
        // the CDeferPos object is destroyed.  All delayed window positions are
        // then done "at once", which can reduce flicker.
        public int MoveWindow(IntPtr hWnd, int x, int y, int nWidth, int nHeight,bool bRepaint)
        {
            uint uFlags = Constants.SWP_NOACTIVATE | Constants.SWP_NOOWNERZORDER | Constants.SWP_NOZORDER;
            if (!bRepaint)
                uFlags |= Constants.SWP_NOREDRAW;
            return User32.SetWindowPos(hWnd, (IntPtr)Constants.HWND_TOP, x, y, nWidth, nHeight, uFlags);
        }

        // SetWindowPos
        // Emulates a call to ::SetWindowPos but the actual call is delayed until
        // the CDeferPos object is destroyed.  All delayed window positions are
        // then done "at once", which can reduce flicker.
        public int SetWindowPos(IntPtr hWnd, IntPtr hWndAfter, int x, int y, int nWidth,int nHeight, uint uFlags)
        {
            if (m_hdwp != IntPtr.Zero)
            {
                return User32.DeferWindowPos(m_hdwp, hWnd, hWndAfter, x, y, nWidth, nHeight,uFlags);
            }
            return -1;
        }
    }
}
