/*
Push-And-Pull Switching: Windows Switching Technique
Copyright (C) 2010 Quan Xu
E-mail:xuquan_sizuo@hotmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using System.Windows.Forms;
using System.Drawing;

using DesktopWindows;

namespace WindowsGroups
{
    //divide windows into cluster/groups
    public class Cluster
    {
        private Desktop m_Desktop = null;
        private static int margin = 4;

        public Cluster()
        {
            m_Desktop = Desktop.DesktopWindows;
        }

        public void DivideCluster(ref ArrayList windowscluster)
        {
            if (m_Desktop != null)
            {
                if (m_Desktop.NumberWindows < 2)
                    return;

                //Windows foreground = m_Desktop.ForegroundWindow;
  
                foreach (Window window in m_Desktop.Values)
                {
                    if (window.IsTopMost || window.WindowStatus == Status.Minimize) // TOPMOST Window and Minimize window
                    {
                        continue;
                    }

                    bool flag = false;
                    if (windowscluster.Count > 0)
                    {
                        foreach (ArrayList cluster in windowscluster)
                        {
                            foreach (Window temp in cluster)
                            {
                                if (m_Desktop.IsIntersection(window, temp))
                                {
                                    cluster.Add(window);
                                    flag = true;
                                    break;
                                }
                            }
                        }
                    }
                    if (!flag)
                    {
                        ArrayList arrtmp = new ArrayList();
                        arrtmp.Add(window);
                        windowscluster.Add(arrtmp);
                    }
                }
            }
        }

        //to support multi-displays
        public void DivideWindowsByScreen(Screen[] screen,ref ArrayList ScreenWindows)
        {
            if (m_Desktop != null)
            {
                foreach (Screen scr in screen)
                {
                    Rectangle rect = scr.Bounds;

                    Rectangle newBounds = new Rectangle(rect.X + margin, rect.Y + margin, rect.Width - 2 * margin, rect.Height - 2*margin); //when windows maximize, the coordinates may not start at(0,0) 
                    ArrayList arrtmp = new ArrayList();

                    foreach (Window window in m_Desktop.Values)
                    {
                        if (window.IsTopMost || window.WindowStatus == Status.Minimize) // TOPMOST Window and Minimize window
                        {
                            continue;
                        }

                        Rectangle objrect = m_Desktop.ChangeWindowtoRectangle(window);
                        if (newBounds.IntersectsWith(objrect))
                        {
                            arrtmp.Add(window);
                        }
                    }
                    ScreenWindows.Add(arrtmp);
                }                
            }
        }

    }
}
