/*
Push-And-Pull Switching: Windows Switching Technique
Copyright (C) 2010 Quan Xu
E-mail:xuquan_sizuo@hotmail.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

using System;
using System.Collections.Generic;
using System.Text;

using System.Diagnostics;
using System.Windows.Forms;
using System.Windows;
using System.Drawing;
using System.Collections;

using Common;
using DesktopWindows;
using Algorithm;

namespace WindowsGroups
{
    //windows group
    public class WindowsGroup
    {
        private Desktop m_Desktop;

        public WindowsGroup()
        {
            m_Desktop = Desktop.DesktopWindows;
        }

        private Rectangle ChangeWindowtoRectangle(Window window)
        {
            return new Rectangle(window.Left, window.Top, window.Width, window.Height);
        }

        private bool IsIntersection(Window window1, Window window2)
        {
            Rectangle rect1 = ChangeWindowtoRectangle(window1);
            Rectangle rect2 = ChangeWindowtoRectangle(window2);

            return rect1.IntersectsWith(rect2);
        }

        //when the intersection region is small, we consider these two windows are not intersection
        private bool IsIntersection(Window window1, Window window2, double ratio)
        {
            Rectangle rect1 = ChangeWindowtoRectangle(window1);
            Rectangle rect2 = ChangeWindowtoRectangle(window2);

            Rectangle region = Rectangle.Intersect(rect1, rect2);
            if (region != null)
            {
                int area1 = rect1.Height * rect1.Width;
                int area2 = rect2.Height * rect2.Width;
                if (area1 != 0 && area2 != 0)
                {
                    float r1 = (float)(region.Height * region.Width) / area1;
                    float r2 = (float)(region.Height * region.Width) / area2;

                    if (r1 < ratio && r2 < ratio)
                    {
                        return false;
                    }
                    else
                    {
                        return true;
                    }
                }
            }
            return false;
        }

        private bool IsIntersection(Window window1, Window window2, out Rectangle intersectionregion)
        {
            Rectangle rect1 = ChangeWindowtoRectangle(window1);
            Rectangle rect2 = ChangeWindowtoRectangle(window2);

            intersectionregion = Rectangle.Intersect(rect1, rect2);
            return rect1.IntersectsWith(rect2);
        }

        private void DivideGroup(Window operatorWindow, ArrayList inputWindows, ArrayList windowsGroup, ArrayList windowHandles, double overlappingratio)
        {
            windowsGroup.Add(operatorWindow);
            windowHandles.Add(operatorWindow.Handle);

            uint zOrder = operatorWindow.ZOrder;

            if (zOrder >= m_Desktop.Count - 1)
                return;

            if (operatorWindow == null)
                return;

            if (inputWindows.Count < 0)
                return;

            bool flag = false;

            foreach (Window window in inputWindows)
            {
                if (window.ZOrder > zOrder)
                {
                    flag = false;

                    ArrayList NeedJudgeSet = new ArrayList();
                    OriginalNeedJudgeIntersectSet(operatorWindow, window, inputWindows, NeedJudgeSet);
                    double totalintersectionregionarea = 0;
                    int opwindowarea = window.Area;

                    //2009-09-23
                    LineSegmentTree segtree = new LineSegmentTree();
                    ArrayList rectSet = new ArrayList();

                    foreach (Window needJudgeWindow in NeedJudgeSet)
                    {
                        //2009-09-13
                        Rectangle intersection = new Rectangle();

                        if (IsIntersection(needJudgeWindow, window, out intersection))
                        {
                            //2009-09-23 fix, using line segment tree to calculate union area
                            rectSet.Add(intersection);
                            totalintersectionregionarea = segtree.NRectangleUnion(rectSet);
                            if (opwindowarea > 0)
                            {
                                if (totalintersectionregionarea / opwindowarea > overlappingratio)
                                {
                                    flag = true;
                                    break;
                                }
                            }
                        }
                    }
                    if (!flag)
                    {
                        windowsGroup.Add(window);
                        windowHandles.Add(window.Handle);
                    }
                }
            }
        }

        //06/10/2009 
        private void RemnantWindows(ArrayList doneHandles, ArrayList inputWindows, ArrayList remnantWindows)
        {
            if (doneHandles.Count > 0)
            {
                bool flag = false;
                foreach (Window window in inputWindows)
                {
                    flag = false;
                    foreach (IntPtr handle in doneHandles)
                    {
                        if (handle == window.Handle)
                        {
                            flag = true;
                            break;    
                        }
                    }
                    if (!flag)
                    {
                        remnantWindows.Add(window);
                    }
                }
            }
        }

        private void OriginalNeedJudgeIntersectSet(Window oriWindow, Window NeedJudgeWindow, ArrayList inputWindows, ArrayList NeedJudgeSet)
        {
            if (oriWindow == null || NeedJudgeWindow == null)
                return;

            foreach (Window window in inputWindows)
            {
                if (window.ZOrder < NeedJudgeWindow.ZOrder && window.ZOrder >= oriWindow.ZOrder)
                {
                    NeedJudgeSet.Add(window);
                }
            }
        }

        public void DivideGroups(Window operatorWindow, ArrayList Groups, double overlappingratio)
        {
            if ((m_Desktop.NumberWindows < 2) || (operatorWindow == null))
                return;

            ArrayList inputWindows = new ArrayList();
            foreach (Window window in m_Desktop.Values)
            {
                if ((window.ZOrder >= operatorWindow.ZOrder) && (window.WindowStatus != Status.Minimize) && (!window.IsTopMost))
                {
                    if (window.Area > 0) //skype has two windows, one of them the size is zero 
                    {
                        inputWindows.Add(window); 
                    }
                }
            }

            int number = 0;
            Window start = operatorWindow;

            while (number < m_Desktop.NumberWindows)
            {
                ArrayList windowsGroup = new ArrayList();
                ArrayList windowHandles = new ArrayList();
                ArrayList remnantWindows = new ArrayList();

                DivideGroup(start, inputWindows, windowsGroup, windowHandles, overlappingratio);
                Groups.Add(windowsGroup);
                RemnantWindows(windowHandles, inputWindows, remnantWindows);
                number += windowsGroup.Count;

                //input for next step 
                if (remnantWindows.Count > 0)
                {
                    start = (Window)remnantWindows[0];
                    inputWindows.Clear();
                    foreach (Window window in remnantWindows)
                    {
                        inputWindows.Add(window);
                    }
                }
                else
                    return;
            }
        }

        public void DivideGroups(Window operatorWindow, ArrayList oriWindows, ArrayList resGroups, double overlappingratio)
        {
            if ((oriWindows.Count < 2) || (operatorWindow == null))
                return;

            ArrayList inputWindows = new ArrayList();
            foreach (Window window in oriWindows)
            {
                if ((window.ZOrder >= operatorWindow.ZOrder) && (window.WindowStatus != Status.Minimize))
                {
                    if (window.Area > 0) //skype has two windows, one of them the size is zero 
                    {
                        inputWindows.Add(window); 
                    }
                }
            }

            int number = 0;
            Window start = operatorWindow;

            while (number < oriWindows.Count)
            {
                ArrayList windowsGroup = new ArrayList();
                ArrayList windowHandles = new ArrayList();
                ArrayList remnantWindows = new ArrayList();

                DivideGroup(start, inputWindows, windowsGroup, windowHandles, overlappingratio);
                resGroups.Add(windowsGroup);
                RemnantWindows(windowHandles, inputWindows, remnantWindows);
                number += windowsGroup.Count;

                //input for next step 
                if (remnantWindows.Count > 0)
                {
                    start = (Window)remnantWindows[0];
                    inputWindows.Clear();
                    foreach (Window window in remnantWindows)
                    {
                        inputWindows.Add(window);
                    }
                }
                else
                    return;
            }
        }

        public int GetDefinitionActiveWindowGroupIndex(IntPtr definitionActiveWindowHandle, ArrayList Layers)
        {
            int index = -1;
            if (Layers.Count > 0)
            {
                foreach (ArrayList windowsList in Layers)
                {
                    ++index;
                    foreach (Window window in windowsList)
                    {
                        if (window.Handle == definitionActiveWindowHandle)
                        {
                            return index;
                        }
                    }
                }
                return -1;
            }
            else
                return -1;
        }

        public void SwitchGroup(int index1, int index2, ref ArrayList Groups)
        {
            //index valid
            int numberGroups = Groups.Count;
            Debug.Assert(index1 >= 0 && index1 < numberGroups);
            Debug.Assert(index2 >= 0 && index2 < numberGroups);

            ArrayList temp = new ArrayList();
            temp = (ArrayList)Groups[index2];
            Groups[index2] = Groups[index1];
            Groups[index1] = temp;
        }

        public void SwitchWindow(int index1, int index2, ref ArrayList Group)
        {
            //index valid
            int numberGroup = Group.Count;
            Debug.Assert(index1 >= 0 && index1 < numberGroup);
            Debug.Assert(index2 >= 0 && index2 < numberGroup);

            Window temp = (Window)Group[index2];
            Group[index2] = Group[index1];
            Group[index1] = temp;
        }

        //n windows intersection region
        public int NWindowsIntersectionRegionUnion(ArrayList opWindowsSet)
        {
            int nCount = opWindowsSet.Count;
            int area = 0;
            if (nCount > 0)
            {
                ArrayList rect = new ArrayList();
                foreach (Window tmp in opWindowsSet)
                {
                    Rectangle tmprect = ChangeWindowtoRectangle(tmp);
                    rect.Add(tmprect);
                }
                LineSegmentTree segtree = new LineSegmentTree();
                area = segtree.NRectangleUnion(rect);
            }
            return area;
        }
    }
}
